import pandas as pd
from scipy import stats
from scipy.stats import normaltest, chi2_contingency
from scikit_posthocs import posthoc_dunn
from statsmodels.stats.anova import AnovaRM
from statsmodels.stats.multicomp import pairwise_tukeyhsd
from statsmodels.stats.proportion import proportions_ztest
from pingouin import pairwise_gameshowell, welch_anova
from itertools import combinations
from typing import List, Dict, Union, Any


class PosthocResult:
    def __init__(
        self, result_pretty_text: str, p_value: float, groups: List["GroupResult"]
    ):
        self.result_pretty_text: str = result_pretty_text
        self.p_value: float = p_value
        self.groups: List[GroupResult] = groups


class GroupResult:
    def __init__(
        self,
        name: str,
        sample_size: int,
        mean: float,
        median: Union[float, None] = None,
    ):
        self.name: str = name
        self.sample_size: int = sample_size
        self.mean: float = mean
        self.median: Union[float, None] = median


class PosthocResults:
    def __init__(self):
        self.significant_results: List[PosthocResult] = []

    def add_result(
        self, result_pretty_text: str, p_value: float, groups_info: List[Dict[str, Any]]
    ):
        groups: List[GroupResult] = [GroupResult(**group) for group in groups_info]
        result: PosthocResult = PosthocResult(result_pretty_text, p_value, groups)
        self.significant_results.append(result)


def is_dichotomous(df: pd.DataFrame, variable: str) -> bool:
    # Check if the variable has only two unique values
    return len(df[variable].unique()) == 2


def is_numeric(df: pd.DataFrame, variable: str) -> bool:
    # Check if the variable is of numeric type
    return df[variable].dtype in [float, int]


def is_gaussian(
    df: pd.DataFrame, group_column: str, variable: str, p_value_threshold: float
) -> bool:
    # Iterate over each unique group in the specified column
    for group in df[group_column].unique():
        # Subset the dataframe for the current group and test for normality
        group_data = df[df[group_column] == group][variable]
        _, p_value = normaltest(group_data)

        # If the p-value is less than the threshold, the data is not Gaussian
        if p_value < p_value_threshold:
            return False  # Early return if any group is not Gaussian

    # If all groups passed the normality test, return True
    return True


def get_chi_square_significance(
    df: pd.DataFrame, group_column: str, variable: str
) -> float:
    contingency_table = pd.crosstab(df[group_column], df[variable])
    _, pvalue, _, _ = chi2_contingency(contingency_table)
    return pvalue


def get_chi_square_posthoc_results(
    df: pd.DataFrame, group_column: str, variable: str, p_value_threshold: float
) -> PosthocResults:
    posthoc_results = PosthocResults()
    contingency_table = pd.crosstab(df[group_column], df[variable])
    group_combinations = [
        (group_a, group_b)
        for i, group_a in enumerate(contingency_table.index)
        for group_b in contingency_table.index[i + 1 :]
    ]
    p_values: List[float] = []

    for group_a, group_b in group_combinations:
        if 1 not in contingency_table.columns:
            continue
        count = [contingency_table.loc[group_a, 1], contingency_table.loc[group_b, 1]]
        nobs = [
            contingency_table.loc[group_a].sum(),
            contingency_table.loc[group_b].sum(),
        ]
        rate_a = (count[0] / nobs[0]) * 100
        rate_b = (count[1] / nobs[1]) * 100

        _, p_value = proportions_ztest(count, nobs)
        p_values.append(p_value)

        groups_info = sorted(
            [
                {"name": group_a, "sample_size": nobs[0], "mean": rate_a},
                {"name": group_b, "sample_size": nobs[1], "mean": rate_b},
            ],
            key=lambda x: x["mean"],
            reverse=True,
        )

        if p_value < p_value_threshold:
            result_pretty_text = f"{groups_info[0]['name']} ({groups_info[0]['mean']:.1f}%) vs {groups_info[1]['name']} ({groups_info[1]['mean']:.1f}%) (p={p_value:.3f})"
            posthoc_results.add_result(result_pretty_text, p_value, groups_info)

    if len(p_values) == 0:
        return posthoc_results
    # Apply Bonferroni correction
    corrected_alpha = p_value_threshold / len(p_values)
    posthoc_results.significant_results = [
        result
        for result in posthoc_results.significant_results
        if result.p_value < corrected_alpha
    ]

    return posthoc_results


def is_levene_significant(
    df: pd.DataFrame, group_column: str, variable: str, p_value_threshold: float
) -> bool:
    # Prepare data for homogeneity of variances test
    groups = df[group_column].unique()
    data_groups = [df[df[group_column] == group][variable] for group in groups]

    # Perform Levene's test for equal variances
    _, p_levene = stats.levene(*data_groups)

    # Check if variances are equal
    return p_levene < p_value_threshold


def get_welch_anova_significance(
    df: pd.DataFrame, group_column: str, variable: str
) -> float:
    # Perform Welch's ANOVA
    welch_anova_results = welch_anova(data=df, dv=variable, between=group_column)
    # Check if the p-value from Welch's ANOVA is significant
    return welch_anova_results.loc[0, "p-unc"]


def get_games_howell_posthoc_results(
    df: pd.DataFrame, group_column: str, variable: str, p_value_threshold: float
) -> PosthocResults:
    posthoc_results = PosthocResults()
    games_howell_results = pairwise_gameshowell(
        data=df, dv=variable, between=group_column
    )
    group_means = df.groupby(group_column)[variable].mean()
    group_medians = df.groupby(group_column)[variable].median()
    group_sizes = df.groupby(group_column).size()

    for _, row in games_howell_results.iterrows():
        p_value = row["pval"]
        if p_value < p_value_threshold:
            group_a, group_b = row["A"], row["B"]
            mean_a, mean_b = group_means[group_a], group_means[group_b]
            median_a, median_b = group_medians[group_a], group_medians[group_b]
            size_a, size_b = group_sizes[group_a], group_sizes[group_b]

            # Order groups so the one with the higher mean is first
            groups_info = sorted(
                [
                    {
                        "name": group_a,
                        "sample_size": size_a,
                        "mean": mean_a,
                        "median": median_a,
                    },
                    {
                        "name": group_b,
                        "sample_size": size_b,
                        "mean": mean_b,
                        "median": median_b,
                    },
                ],
                key=lambda x: x["mean"],
                reverse=True,
            )

            result_pretty_text = f"{groups_info[0]['name']} ({groups_info[0]['mean']:.2f}) vs {groups_info[1]['name']} ({groups_info[1]['mean']:.2f}) (p={p_value:.3f})"
            posthoc_results.add_result(result_pretty_text, p_value, groups_info)

    return posthoc_results


def get_anova_results(df: pd.DataFrame, group_column: str, variable: str) -> AnovaRM:
    # Perform one-way ANOVA
    anova_results = AnovaRM(
        data=df, depvar=variable, subject="session_id", within=[group_column]
    ).fit()
    return anova_results


def get_oneway_anova_significance(
    df: pd.DataFrame, group_column: str, variable: str
) -> float:
    group_names = df[group_column].unique()
    # Perform one-way ANOVA
    group_data = [df[df[group_column] == group][variable] for group in group_names]
    _, pvalue = stats.f_oneway(*group_data)
    return pvalue


def get_crushal_wallis_significance(
    df: pd.DataFrame, group_column: str, variable: str
) -> float:
    group_names = df[group_column].unique()
    # Perform Kruskal-Wallis test
    group_data = [df[df[group_column] == group][variable] for group in group_names]
    _, pvalue = stats.kruskal(*group_data)
    return pvalue


def get_tukeyhsd_posthoc_results(
    df: pd.DataFrame, group_column: str, variable: str, p_value_threshold: float
) -> PosthocResults:
    posthoc_results = PosthocResults()
    # Perform Tukey's HSD test
    tukey = pairwise_tukeyhsd(
        endog=df[variable],
        groups=df[group_column],
        alpha=p_value_threshold,
    )

    # Calculate means and medians for each group
    group_means = df.groupby(group_column)[variable].mean()
    group_medians = df.groupby(group_column)[variable].median()
    group_sizes = df.groupby(group_column).size()

    # Process only significant results
    significant_results = tukey._results_df[tukey._results_df["reject"] == True]
    for _, row in significant_results.iterrows():
        group_a, group_b = row["group1"], row["group2"]
        p_value = row["p-adj"]
        mean_a, mean_b = group_means[group_a], group_means[group_b]
        median_a, median_b = group_medians[group_a], group_medians[group_b]
        size_a, size_b = group_sizes[group_a], group_sizes[group_b]

        # Order groups so the one with the higher mean is first
        groups_info = sorted(
            [
                {
                    "name": group_a,
                    "sample_size": size_a,
                    "mean": mean_a,
                    "median": median_a,
                },
                {
                    "name": group_b,
                    "sample_size": size_b,
                    "mean": mean_b,
                    "median": median_b,
                },
            ],
            key=lambda x: x["mean"],
            reverse=True,
        )

        result_pretty_text = f"{groups_info[0]['name']} ({groups_info[0]['mean']:.2f}) vs {groups_info[1]['name']} ({groups_info[1]['mean']:.2f}) (p={p_value:.3f})"
        posthoc_results.add_result(result_pretty_text, p_value, groups_info)

    return posthoc_results


def get_dunn_posthoc_results(df, group_column, variable, p_value_threshold):
    posthoc_results = PosthocResults()
    # Perform Dunn's test
    dunn_results = posthoc_dunn(
        df,
        val_col=variable,
        group_col=group_column,
        p_adjust="holm",
    )
    group_means = df.groupby(group_column)[variable].mean()
    group_medians = df.groupby(group_column)[variable].median()
    group_sizes = df.groupby(group_column).size()

    for group_a, group_b in combinations(dunn_results.index, 2):
        p_value = dunn_results.at[group_a, group_b]
        if p_value < p_value_threshold:
            mean_a, mean_b = group_means[group_a], group_means[group_b]
            median_a, median_b = group_medians[group_a], group_medians[group_b]
            size_a, size_b = group_sizes[group_a], group_sizes[group_b]

            # Order groups so the one with the higher mean is first
            groups_info = sorted(
                [
                    {
                        "name": group_a,
                        "sample_size": size_a,
                        "mean": mean_a,
                        "median": median_a,
                    },
                    {
                        "name": group_b,
                        "sample_size": size_b,
                        "mean": mean_b,
                        "median": median_b,
                    },
                ],
                key=lambda x: x["mean"],
                reverse=True,
            )

            result_pretty_text = f"{groups_info[0]['name']} ({groups_info[0]['mean']:.2f}) vs {groups_info[1]['name']} ({groups_info[1]['mean']:.2f}) (p={p_value:.3f})"
            posthoc_results.add_result(result_pretty_text, p_value, groups_info)
    return posthoc_results



import pandas as pd
from .significance_tests import (
    is_dichotomous,
    is_numeric,
    is_levene_significant,
    is_gaussian,
    get_chi_square_significance,
    get_chi_square_posthoc_results,
    get_welch_anova_significance,
    get_oneway_anova_significance,
    get_crushal_wallis_significance,
    get_games_howell_posthoc_results,
    get_tukeyhsd_posthoc_results,
    get_dunn_posthoc_results,
)
from typing import Optional, List


class AnalysisResults:
    def __init__(
        self,
        significant_results=None,
        info=None,
        sample_size=None,
        dichotomous_flag=None,
        levene_flag=None,
        gaussian_flag=None,
    ):
        self.significant_results = (
            significant_results if significant_results is not None else []
        )
        self.info = info
        self.sample_size = sample_size
        self.dichotomous_flag = dichotomous_flag
        self.levene_flag = levene_flag
        self.gaussian_flag = gaussian_flag


def get_results(
    df: pd.DataFrame,
    variable_to_analyze: str,
    group_column: str,
    p_value_threshold: float = 0.05,
    min_sample_size: int = 25,
) -> Optional[AnalysisResults]:
    """
    Returns a text with the results of the statistical significance tests.

    Parameters:
    df (pandas.DataFrame): The dataframe containing the data to analyze.
    variable_to_analyze (str): The column name in df that is to be analyzed for statistical significance.
    group_column (str): The name of the column in df that contains the grouping variable.
    p_value_threshold (float, optional): The threshold for determining statistical significance. Defaults to 0.05.
    min_sample_size (int, optional): The minimum sample size for the data. Defaults to 25.

    Returns:
    AnalysisResults: An instance of the AnalysisResults class containing the following attributes:
        - significant_results: A list of results of the statistical significance tests.
        - info: A string containing information about the data if no results were found.
        - sample_size: The sample size of the data.
        - dichotomous_flag: A boolean flag indicating if the data is dichotomous (e.g. boolean).
        - levene_flag: A boolean flag indicating if Levene's test for homogeneity of variances is significant.
        - gaussian_flag: A boolean flag indicating if the data has a Gaussian distribution.
    """
    if p_value_threshold > 0.05:
        print(
            "Warning: p_value_threshold is set to a value higher than the conventional alpha level of 0.05."
        )

    try:
        df[variable_to_analyze]
        df[group_column]
    except KeyError as e:
        print(f"Column not found in DataFrame: {e}")
        return None

    df = df.dropna(subset=[variable_to_analyze])
    sample_size = len(df)
    pvalue = None
    info = None
    dichotomous_flag = False
    levene_flag = False
    gaussian_flag = False
    # Check if any group has less than the minimum sample size
    if (
        df.groupby(group_column)[variable_to_analyze]
        .apply(lambda x: len(x) < min_sample_size)
        .any()
    ):
        info = "not enough data"
    # Check if the variable is dichotomous (e.g. boolean)
    elif dichotomous_flag := is_dichotomous(df, variable_to_analyze):
        # Perform chi-square test
        pvalue = get_chi_square_significance(df, group_column, variable_to_analyze)
        if pvalue < p_value_threshold:
            results = get_chi_square_posthoc_results(
                df, group_column, variable_to_analyze, p_value_threshold
            )
    elif is_numeric(df, variable_to_analyze):
        # Check homogeneity of variances. If significant, use non-parametric tests
        if levene_flag := is_levene_significant(
            df, group_column, variable_to_analyze, p_value_threshold
        ):
            pvalue = get_welch_anova_significance(df, group_column, variable_to_analyze)
            if pvalue < p_value_threshold:
                results = get_games_howell_posthoc_results(
                    df, group_column, variable_to_analyze, p_value_threshold
                )
        # Check if the data has Gaussian distribution, if not use non-parametric tests
        elif gaussian_flag := is_gaussian(
            df, group_column, variable_to_analyze, p_value_threshold
        ):
            pvalue = get_oneway_anova_significance(
                df, group_column, variable_to_analyze
            )
            if pvalue < p_value_threshold:
                results = get_tukeyhsd_posthoc_results(
                    df, group_column, variable_to_analyze, p_value_threshold
                )
        else:
            pvalue = get_crushal_wallis_significance(
                df, group_column, variable_to_analyze
            )
            if pvalue < p_value_threshold:
                results = get_dunn_posthoc_results(
                    df, group_column, variable_to_analyze, p_value_threshold
                )
    else:
        print("Variable has to be dichotomous (e.g. boolean) or numeric")
        info = "Variable has to be dichotomous (e.g. boolean) or numeric"

    # If no results were found, return a message
    if not "results" in locals() or len(results.significant_results) == 0:
        if pvalue is not None:
            info = f"not significant (p={pvalue:.3f})"
    # If results were found, return them
    if "results" in locals():
        significant_results = results.significant_results
    else:
        significant_results = []

    return AnalysisResults(
        significant_results,
        info,
        sample_size,
        dichotomous_flag,
        levene_flag,
        gaussian_flag,
    )


def get_results_pretty_text_header(identifiers: Optional[List[str]] = None) -> str:
    """
    Returns a text with the results of the statistical significance tests.

    Parameters:
    identifiers (list of str, optional): A list of strings that are to be used for identifying the data. Defaults to [].

    Returns:
    str: A formatted text string with the header of the pretty text table.
    """
    identifier_string = ""
    identifier_string_empty = ""
    if identifiers:
        identifier_string = "    ".join(identifiers)
        identifier_string += "    "
        identifier_string_empty = "    ".join([""] * len(identifiers))
        identifier_string_empty += "    "
        output_string = f"{'Identifier':<{len(identifier_string)}}{'n':>10}    {'Levene':<6}    {'Gaussian':<8}    {'Result'}"
    else:
        output_string = f"{'n':>10}    {'Levene':<6}    {'Gaussian':<8}    {'Result'}"

    return output_string


def get_results_pretty_text(
    df: pd.DataFrame,
    variable_to_analyze: str,
    group_column: str,
    p_value_threshold: float = 0.05,
    min_sample_size: int = 25,
    header: bool = True,
    identifiers: Optional[List[str]] = None,
    show_only_significant_results: bool = False,
):
    """
    Returns a text with the results of the statistical significance tests.

    Parameters:
    df (pandas.DataFrame): The dataframe containing the data to analyze.
    variable_to_analyze (str): The column name in df that is to be analyzed for statistical significance.
    group_column (str): The name of the column in df that contains the grouping variable.
    p_value_threshold (float, optional): The threshold for determining statistical significance. Defaults to 0.05.
    min_sample_size (int, optional): The minimum sample size for the data. Defaults to 25.
    header (bool, optional): A boolean flag indicating if the header of the pretty text table should be included. Defaults to True.
    identifiers (list of str, optional): A list of strings that are to be used for identifying the data. Defaults to [].
    show_only_significant_results (bool, optional): A boolean flag indicating if only significant results should be shown. Defaults to False.

    Returns:
    str: A formatted text string with the results of the statistical significance tests.
    """

    analysis_results = get_results(
        df, variable_to_analyze, group_column, p_value_threshold, min_sample_size
    )
    significant_results = analysis_results.significant_results
    info = analysis_results.info
    sample_size = analysis_results.sample_size
    levene_flag = analysis_results.levene_flag
    gaussian_flag = analysis_results.gaussian_flag

    identifier_string = ""
    identifier_string_empty = ""
    if identifiers:
        identifier_string = "    ".join(identifiers)
        identifier_string += "    "
        identifier_string_empty = "    ".join([""] * len(identifiers))
        identifier_string_empty += "    "

    output_string = ""
    if header:
        output_string += get_results_pretty_text_header(identifiers)
        output_string += "\n"

    levene_output = "X" if levene_flag else ""
    gaussian_output = "X" if gaussian_flag else ""

    output_results = []
    if len(significant_results) > 0:
        for i, result in enumerate(significant_results):
            if i == 0:
                output_results.append(
                    f"{identifier_string}{sample_size:>10}    {levene_output:<6}    {gaussian_output:<8}    {result.result_pretty_text}"
                )
            else:
                output_results.append(
                    f"{len(identifier_string) * ' '}{sample_size:>10}    {levene_output:<6}    {gaussian_output:<8}    {result.result_pretty_text}"
                )
    elif not show_only_significant_results:
        output_results.append(
            f"{identifier_string}{sample_size:>10}    {levene_output:<6}    {gaussian_output:<8}    {info}"
        )
    else:
        output_results.append(
            f"{identifier_string}{sample_size:>10}    {levene_output:<6}    {gaussian_output:<8}"
        )
    output_string += "\n".join(output_results)
    return output_string
